package edu.uci.ics.pattis.ics23.collections;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;



//add         : O(N)
//addAll      : O(MN) (M values in Iterable/Collection) 
//remove      : O(N)
//removeAll   : O(MN) (M values in Iterable/Collection)
//retainAll   : O(MN) (M values in Iterable/Collection)
//clear       : O(1)   with garbage collection issue
//contains    : O(N)
//containsAll : O(MN) (M values in Iterable/Collection)
//isEmpty     : O(1)
//size        : O(1)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(N)
//toArray     : O(N)
//newEmpty    : O(1)
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N^2)
//hashCode    : O(N)
//
//Special "index" operations
//add         : O(N)
//remove      : O(N)
//set         : O(1)
//addAll      : O(M+N)
//get         : O(1)
//indexOf     : O(N)
//lastIndexOf : O(N)
//subList     : O(M) (M values from startIndex to StopIndex)



public class ArrayList<E> extends AbstractList<E> implements List<E> {

	
  public ArrayList ()
  {this(1);}
 
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayList (int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayList("+initialCapacity+"): illegal initialCapacity");
 
    list = (E[]) new Object[initialCapacity];
  }

  
  public ArrayList (Iterable<? extends E> es) {
    this(1);
    addAll(es);
  }

  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayList (E[] es) {
    list = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  }
  
  
  
  //This method is required for a concrete class implementing List
  public boolean add (E e) {
    ensureCapacity(objectCount+1);
    list[objectCount++] = e;
    modCount++;
    return true;
  }
	
	
	
  //This method improves the performance of the method inherited
  //  from AbstractList
  public void clear() {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  list[i] = null;
    objectCount = 0;
    modCount++;
  }
	
	
  //This method is required for a concrete class implementing List
  public boolean add (int index, E e) throws IndexOutOfBoundsException {
    if (index < 0 || index > size())
      throw new IndexOutOfBoundsException
        ("ArrayList.add: index("+index+") illegal for size("+ size()+")");

    ensureCapacity(objectCount+1);
    for (int i=objectCount; i>index; i--)
      list[i] = list[i-1];
    list[index] = e;
    objectCount++;
    modCount++;
    return true;
  }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public boolean addAll (int index, Iterable<? extends E> es) throws IndexOutOfBoundsException {
    if (index < 0 || index > size())
      throw new IndexOutOfBoundsException
        ("ArrayList.addAll: index("+index+") illegal for size("+ size()+")");

    ArrayList<E> toAdd = new ArrayList<E>(es);
    int         addSize = toAdd.size();

    if (addSize == 0)
      return false;

    ensureCapacity(objectCount+addSize);

    for (int i=objectCount+addSize-1; i-addSize>=index; i--)
      list[i] = list[i-addSize];
    for (int i=0; i<addSize; i++)
      list[i+index] = toAdd.list[i];
    
    objectCount += addSize;
    modCount++;
    return true;
  }
  
  

  //This method is required for a concrete class implementing List
  public List<E> newEmpty (){
    return new ArrayList<E>();
  }

  
  
  //This method is required for a concrete class implementing List
  public List<E> shallowCopy (){
    //Must construct object and set all instance variables appropriately
    ArrayList<E> answer = new ArrayList<E>(list.length);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.list[i] = list[i];
    return answer;
  }

  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public E remove (int index) throws IndexOutOfBoundsException {
    if (index < 0 || index >= size())
      throw new IndexOutOfBoundsException
        ("ArrayList.remove: index("+index+") illegal for size("+ size()+")");

    E oldValue = list[index];
    for (int i=index; i<objectCount-1; i++)
      list[i] = list[i+1];
 
    objectCount--;
    list[objectCount] = null;
    modCount++;
    return oldValue;
  }
  
  
  
 //This method is required for a concrete class implementing List
  public E set (int index, E e) throws IndexOutOfBoundsException {
    if (index < 0 || index >= size())
      throw new IndexOutOfBoundsException
        ("ArrayList.set: index("+index+") illegal for size("+ size()+")");

    E oldValue = list[index];
    list[index] = e;
    modCount++;
    return oldValue;
  }

  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public E get (int index){
    if (index < 0 || index >= size())
      throw new IndexOutOfBoundsException
        ("ArrayList.get: index("+index+") illegal for size("+ size()+")");
 
    return list[index];	
  }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public int indexOf (E e){
    if (e == null) {
      for (int i=0; i<objectCount; i++)
        if (list[i] == null)
    	  return i;
    }else{
      for (int i=0; i<objectCount; i++)
        if (e.equals(list[i]))
    	  return i;
    }
  	
    return -1;
  }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public int lastIndexOf(E e){
    if (e == null) {
      for (int i=objectCount-1; i>=0; i--)
    	if (list[i] == null)
    	  return i;
    }else{
      for (int i=objectCount-1; i>=0; i--)
   	if (e.equals(list[i]))
    	  return i;
    }
  	
    return -1;
  }

  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public  List<E> subList (int startIndex, int stopIndex) throws IndexOutOfBoundsException {
    if (startIndex > stopIndex)
      return new ArrayList<E>();
    
    if (startIndex < 0 || startIndex >= size() ||
        stopIndex  < 0 || stopIndex  >= size() )
      throw new IndexOutOfBoundsException
        ("ArrayList.subList: startIndex("+startIndex+
         ") and/or stopIndex("+stopIndex+") illegal for size("+size()+")");

    ArrayList<E> answer = new ArrayList<E>(Math.max(0,stopIndex-startIndex+1));
    answer.objectCount = answer.list.length;

    for (int i=0; i<answer.objectCount; i++)
      answer.list[i] = list[i+startIndex];
    
    return answer;
  }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractList
  public int size ()
  {return objectCount;}

  
  
  public Iterator<E> iterator ()
  {return new ArrayListIterator();}
 
  
 
  //Could use the inherited iterator version, adding extra data structure
  //  information
  //public String toString()
  //{return "ArrayList["+objectCount+"/"+list.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayList["+objectCount+"/"+list.length+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+i+"]=" +
                     (list[i] == this ? "(this Collection)" : String.valueOf(list[i])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = list.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      E[] old = list;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
	 newCapacity = minCapacity;
      list = (E[])new Object[newCapacity];
      for (int i=0; i<objectCount; i++)
	list[i] = old[i];
    }
  } 
  


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = list.length;
    if (objectCount < oldCapacity) {
      modCount++;
      E[] old = list;
      list = (E[])new Object[objectCount];
      for (int i=0; i<objectCount; i++)
        list[i] = old[i];
    }
  }
  
  
  

  //Concrete data structure for implementing a List via an array
  //The order of the values in the array IS important
  private E[] list;
  private int objectCount;


 
 
  private class ArrayListIterator implements Iterator<E> {

    public boolean hasNext()
    {return nextIteratorIndex < objectCount;}

    
    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayListIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayListIterator.next");

      E answer = list[nextIteratorIndex++];
      removedAlready = false;
      return answer;
    }


    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayListIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayListIterator.remove");

      for (int i=nextIteratorIndex-1; i<objectCount-1; i++)
        list[i] = list[i+1];
      list[--objectCount] = null;
      nextIteratorIndex--;
      expectedModCount     = ++modCount;
      removedAlready       = true;
    }


    private int     nextIteratorIndex  = 0;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }
 
}
