package edu.uci.ics.pattis.ics23.collections;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;



//add         : O(1)
//addAll      : O(MN) (M values in Iterable/Collection) 
//remove      : O(N)
//removeAll   : O(MN) (M values in Iterable/Collection)
//retainAll   : O(MN) (M values in Iterable/Collection)
//clear       : O(N)   with garbage collection issue
//contains    : O(N)
//containsAll : O(MN) (M values in Iterable/Collection)
//isEmpty     : O(N)
//size        : O(N)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(N)
//toArray     : O(N)
//makeEmpty   : O(1)
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N^2)
//hashCode    : O(N)
//
//Special "index" operations
//add         : O(N)
//remove      : O(N)
//set         : O(1)
//addAll      : O(MN^2)
//get         : O(N)
//indexOf     : O(N)
//lastIndexOf : O(N)



public class ArrayListMinimal<E> extends AbstractList<E> implements List<E> {

  
  public ArrayListMinimal ()
  {this(1);}
 

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayListMinimal (int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayList("+initialCapacity+"): illegal initialCapacity");
 
    list = (E[]) new Object[initialCapacity];
  }

  
  public ArrayListMinimal (Iterable<? extends E> es) {
    this(1);
    addAll(es);
  }

  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayListMinimal (E[] es) {
    list = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  }
  
  
  
  //This method is required for a concrete class implementing List
  public boolean add (E e) {
    ensureCapacity(objectCount+1);
    list[objectCount++] = e;
    modCount++;
    return true;
  }
  
  
  
  //This method is required for a concrete class implementing List
  public boolean add (int index, E e) throws IndexOutOfBoundsException {
    if (index < 0 || index > size())
      throw new IndexOutOfBoundsException("ArrayList.add: index = " + index +" but size = " + size());

    ensureCapacity(objectCount+1);
    for (int i=objectCount; i>index; i--)
      list[i] = list[i-1];
    list[index] = e;
    objectCount++;
    modCount++;
    return true;
  }
  
  
  
  //This method is required for a concrete class implementing List
  public List<E> newEmpty (){
    return new ArrayListMinimal<E>();
  }

  
  
  //This method is required for a concrete class implementing List
  public List<E> shallowCopy (){
    //Must construct object and set all instance variables appropriately
    ArrayListMinimal<E> answer = new ArrayListMinimal<E>(list.length);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.list[i] = list[i];
    return answer;
  }

  
  
 //This method is required for a concrete class implementing List
  public E set (int index, E e) throws IndexOutOfBoundsException {
    if (index < 0 || index >= size())
      throw new IndexOutOfBoundsException("ArrayList.set: index("+index+") illegal for size("+ size()+")");

    E oldValue = list[index];
    list[index] = e;
    modCount++;
    return oldValue;
  }

  
  
  //This method is required for a concrete class implementing List
  public Iterator<E> iterator ()
  {return new ArrayListIterator();}
 
  
 
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArrayList["+objectCount+"/"+list.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayList["+objectCount+"/"+list.length+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+i+"]=" +
                     (list[i] == this ? "(this Collection)" : String.valueOf(list[i])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }



  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = list.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      E[] old = list;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
	 newCapacity = minCapacity;
      list = (E[])new Object[newCapacity];
      for (int i=0; i<objectCount; i++)
	list[i] = old[i];
    }
  } 
  


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = list.length;
    if (objectCount < oldCapacity) {
      modCount++;
      E[] old = list;
      list = (E[])new Object[objectCount];
      for (int i=0; i<objectCount; i++)
        list[i] = old[i];
    }
  }
  
  
  

  //Concrete data structure for implementing a List via an array
  //The order of the values in the array IS important
  private E[] list;
  private int objectCount;


 
 
  private class ArrayListIterator implements Iterator<E> {

    public boolean hasNext()
    {return nextIteratorIndex < objectCount;}

    
    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayListIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayListIterator.next");

      E answer = list[nextIteratorIndex++];
      removedAlready = false;
      return answer;
    }


    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayListIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayListIterator.remove");

      for (int i=nextIteratorIndex-1; i<objectCount-1; i++)
        list[i] = list[i+1];
      list[--objectCount] = null;
      nextIteratorIndex--;
      expectedModCount     = ++modCount;
      removedAlready       = true;
    }


    private int     nextIteratorIndex  = 0;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }
 
}
