package edu.uci.ics.pattis.ics23.collections;


import java.util.Comparator;
import java.util.Iterator;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;



//add         : O(N)
//addAll      : O(M) (M values in Iterable), amortized
//remove      : O(1)
//clear       : O(1) with garbage collection issue
//peek        : O(1)
//isEmpty     : O(1)
//size        : O(1)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(N)
//toArray     : O(N)
//newEmpty    : O(1)
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N)
//hashCode    : O(N)



public class ArrayPriorityQueue<E> extends AbstractPriorityQueue<E> implements PriorityQueue <E> {


  public ArrayPriorityQueue (Comparator<? super E> priorityComparator) throws IllegalArgumentException
  {this(1, priorityComparator);}
  

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayPriorityQueue (int                   initialCapacity,
                             Comparator<? super E> priorityComparator) throws IllegalArgumentException {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayPriorityQueue: Illegal initialCapacity: "+  initialCapacity);
    if (priorityComparator == null)
      throw new IllegalArgumentException
        ("ArrayPriorityQueue: null priorityComparator");
 
    this.priorityComparator = priorityComparator;
    pq = (E[]) new Object[initialCapacity];
  }


  public ArrayPriorityQueue (Iterable<? extends E> es, Comparator<? super E> priorityComparator) throws IllegalArgumentException {
   this(1, priorityComparator);
   addAll(es);
  }
  

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayPriorityQueue (E[] es, Comparator<? super E> priorityComparator) throws IllegalArgumentException {
    if (priorityComparator == null)
      throw new IllegalArgumentException
       ("ArrayPriorityQueue: null priorityComparator");

    this.priorityComparator = priorityComparator;
    pq = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  }
 


  //This method is required for a concrete class implementing PriorityQueue
  public boolean add (E e) {
    ensureCapacity(objectCount+1);
    pq[objectCount++] = e;
    for (int i=objectCount-1; i>=1; i--)
      if (priorityComparator.compare(pq[i-1],pq[i]) <= 0 )
        break;
      else {
        E temp = pq[i];
        pq[i]  = pq[i-1];
        pq[i-1]= temp;
      }
    
    modCount++;
    return true;
  }
  
 
  
  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E remove () throws NoSuchElementException {
    if (isEmpty())
      throw new NoSuchElementException
        ("ArrayPriorityQueue.remove: empty queue");

    E answer        = pq[--objectCount];
    pq[objectCount] = null;
    modCount++;
    return answer;
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public void clear () {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  pq[i] = null;
    objectCount = 0;
    modCount++;
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E peek () throws NoSuchElementException {
    if (isEmpty())
      throw new NoSuchElementException
        ("ArrayPriorityQueue.peek: empty queue");

    return pq[objectCount-1];
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public int size ()
  {return objectCount;}
 


  //This method is required for a concrete class implementing PriorityQueue
  public Iterator<E> iterator()
  {return new ArrayPriorityQueueIterator();}
  
 
   
  //This method is required for a concrete class implementing List
  public PriorityQueue<E> newEmpty (){
    return new ArrayPriorityQueue<E>(priorityComparator);
  }

  
  
  //This method is required for a concrete class implementing PriorityQueue
  public PriorityQueue<E> shallowCopy() {
    //Must construct object and set all instance variables appropriately
    ArrayPriorityQueue<E> answer = new ArrayPriorityQueue<E>(pq.length,priorityComparator);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.pq[i] = pq[i];
    return answer;
  }
  
  
  
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArrayPriorityQueue["+objectCount+"/"+pq.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayPriorityQueue["+objectCount+"/"+pq.length+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+i+"]=" +
                     (pq[i] == this ? "(this Collection)" : String.valueOf(pq[i])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }

  
  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = pq.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      Object old[] = pq;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
        newCapacity = minCapacity;
      pq = (E[])new Object[newCapacity];
      System.arraycopy(old,0,pq,0,objectCount);
    }
  } 
  
  
  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = pq.length;
    if (objectCount < oldCapacity) {
      modCount++;
      Object old[] = pq;
      pq = (E[])new Object[objectCount];
      System.arraycopy(old,0,pq,0,objectCount);
    }
  }

   


  //Concrete data structure for implementing a PriorityQueue via an array
  //The order of the values in the array IS important (lowest to highest priority)
  private E[]                   pq;
  private Comparator<? super E> priorityComparator;
  private int                   objectCount = 0;

  
  

  private class ArrayPriorityQueueIterator implements Iterator<E> {

    public boolean hasNext()
    {return nextIteratorIndex != -1;}


    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayPriorityQueueIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayPriorityQueueIterator.next");

      removedAlready = false;
      return pq[nextIteratorIndex--];
    }


    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayPriorityQueueIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayPriorityQueueIterator.remove");

      for (int i=nextIteratorIndex+1; i<objectCount-1; i++)
        pq[i] = pq[i+1];
      pq[--objectCount] = null;
      expectedModCount  = ++modCount;
      removedAlready    = true;
    }


    private int     nextIteratorIndex  = objectCount-1;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }

}
