package edu.uci.ics.pattis.ics23.collections;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;



//add         : O(N)
//addAll      : O(MN) (M values in Iterable/Collection) 
//remove      : O(N)
//removeAl    : O(MN) (M values in Iterable/Collection)
//retainAll   : O(MN) (M values in Iterable/Collection)
//clear       : O(1)   with garbage collection issue
//contains    : O(N)
//containsAll : O(MN) (M values in Iterable/Collection)
//isEmpty     : O(1)
//size        : O(1)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(1)
//toArray     : O(N)
//newEmpty    : O(1)
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N^2)
//hashCode    : O(N)



public class ArraySet<E> extends AbstractSet<E> implements Set<E> {

  public ArraySet()
  {this(1);}
	
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArraySet(int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException("ArraySet("+initialCapacity+"): illegal initialCapacity");

    set = (E[])new Object[initialCapacity];
  }

  public ArraySet (Iterable<? extends E> es) {
    this(1);
    addAll(es);
  }

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArraySet (E[] es) {
    set = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  } 

	
	 
  //This method is required for a concrete class implementing Stack
  public boolean add(E e) {
    if (contains(e))
      return false;
    ensureCapacity(objectCount+1);
    set[objectCount++] = e;
    modCount++;
    return true;
  }

	
	
  //This method improves the performance of the method inherited
  //  from AbstractCollection
  public boolean remove(E e) {
    int i = indexOf(e);
    if (i == -1)
      return false;
  
    return removeAt(i);
  }

	
	
  //This method improves the performance of the method inherited
  //  from AbstractCollection
  public void clear() {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  set[i] = null;
    objectCount = 0;
    modCount++;
  }

	
	
  //This method improves the performance of the method inherited
  //  from AbstractCollection
  public boolean contains(Object o) {
    return indexOf(o) != -1;
  }


	
  //This method is required for a concrete class implementing Stack
  public Iterator<E> iterator() {
    return new ArraySetIterator();
  }

	
	
  //This method improves the performance of the method inherited
  //  from AbstractCollection
  public int size() {
    return objectCount;
  }

	
	
  //This method is required for a concrete class implementing Set
  public Set<E> newEmpty (){
    return new ArraySet<E>();
  }

  
  
  //This method is required for a concrete class implementing Stack
  public Set<E> shallowCopy() {
  //Must construct object and set all instance variables appropriately
    ArraySet<E> answer = new ArraySet<E>(set.length);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.set[i] = set[i];
    return answer;
  }

	
	
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArraySet["+objectCount+"/"+set.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArraySet["+objectCount+"/"+set.length+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+i+"]=" +
                     (set[i] == this ? "(this Collection)" : String.valueOf(set[i])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }


  
  
  private int indexOf(Object o) {
  if (o == null) {
    for (int i = 0; i<objectCount; i++)
      if (set[i] == null)
        return i;
    } else
      for (int i = 0; i<objectCount; i++)
        if (o.equals(set[i]))
          return i;
  	
    return -1;
  }
  
 
  
  private boolean removeAt(int i) {
    set[i]           = set[--objectCount]; 	
    set[objectCount] = null;
    modCount++;
    return true;
  }
  
  
  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = set.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      E[] old = set;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
 	newCapacity = minCapacity;
      set = (E[])new Object[newCapacity];
      for (int i=0; i<objectCount; i++)
        set[i] = old[i];
    }
  } 
   


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = set.length;
    if (objectCount < oldCapacity) {
      modCount++;
      E[] old = set;
      set = (E[])new Object[objectCount];
      for (int i=0; i<objectCount; i++)
        set[i] = old[i];
    }
  }
   
   
   
   
  //Concrete data structure for implementing a Set via an array
  //The order of the values in the array IS NOT important
  private E[] set;
  private int objectCount;

  
  
  
  private class ArraySetIterator implements Iterator<E> {
    
    public boolean hasNext()
    {return nextIteratorIndex < objectCount;}
   
    
    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArraySetIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArraySetIterator.next");
      
      E answer = set[nextIteratorIndex++];
      removedAlready = false;
      return answer;
    }
   
   
    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArraySetIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArraySetIterator.remove");

      removeAt(nextIteratorIndex-1);
      nextIteratorIndex--;
      expectedModCount = modCount;  //removeAt: objectCount-- modCount++
      removedAlready   = true;
   }
    
 
    private int     nextIteratorIndex  = 0;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }

}
